using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using Tao.OpenGl;
using Tao.Platform.Windows;
using SkeletalAnimation;
using Util;

namespace OpenGLFramework {

	public partial class OpenGLForm : Form {
		
		#region Member Variables

		private IntPtr hDC;     // Device context
		private IntPtr hRC;     // Rendering context

		private ArrayList meshes;
		private ArrayList animations;
		
		private Animator animator;
		
		private Camera camera;

		private bool doRenderSkin;
		private bool doRenderWireframe;
		private bool doRenderSkeleton;

		#endregion

		#region Constructor

		public OpenGLForm() {
			InitializeComponent();
			animationTimer.Interval = 1000/fpsTrackBar.Value;
			createRenderContext(32);
			initOpenGL();

			meshes = new ArrayList();
			animations = new ArrayList();
			animator = new Animator();

			// Load the meshes
			loadMesh("../../../male.ram");
			loadMesh("../../../switch.ram");

			// Load the animations
			loadAnimation("../../../walk.raa");
			loadAnimation("../../../wave.raa");

			// Set the current mesh
			animator.setMesh(((MeshData)meshes[0]).createInstance());
			comboBoxMeshes.SelectedIndex = 0;

			// Makes the initial transform
			animator.getMesh().transformVertices();

			// Add the walking animation cycle
			animator.addMergeAnimation((Animation)animations[0], 1.0);
			
			// Renderflags
			doRenderSkin = true;
			doRenderWireframe = false;
			doRenderSkeleton = false;
			
			// Camera code
			camera = new Camera();
			Vector cameraPos = new Vector();
			cameraPos[0] = 0; cameraPos[1] = 0; cameraPos[2] = 18;
			camera.setPosition(cameraPos);
		}

		#endregion

		#region Load Model

		private void loadMesh(String filename) {
			RAMLoader ramLoader = new RAMLoader();
			MeshData newMesh = ramLoader.readRAM(filename);
			comboBoxMeshes.Items.Add(newMesh.getName());
			meshes.Add(newMesh);			
		}

		#endregion

		#region Load Animation

		private void loadAnimation(String filename) {
			RAALoader raaLoader = new RAALoader();
			animations.Add(raaLoader.readRAA(filename));			
		}

		#endregion

		#region Create Rendering Context

		private bool createRenderContext(int bits) {

			GC.Collect();   // Request a garbage collection
            Kernel.SetProcessWorkingSetSize(Process.GetCurrentProcess().Handle, -1, -1);    // Empty the working set of the process

			// The pixelformat
			Gdi.PIXELFORMATDESCRIPTOR pfd = new Gdi.PIXELFORMATDESCRIPTOR();
			pfd.nSize = (short)Marshal.SizeOf(pfd);
			pfd.nVersion = 1;
			pfd.dwFlags = Gdi.PFD_DRAW_TO_WINDOW | Gdi.PFD_SUPPORT_OPENGL | Gdi.PFD_DOUBLEBUFFER;
			pfd.iPixelType = (byte)Gdi.PFD_TYPE_RGBA;
			pfd.cColorBits = (byte)bits;
			pfd.cRedBits = 0;
			pfd.cRedShift = 0;
			pfd.cGreenBits = 0;
			pfd.cGreenShift = 0;
			pfd.cBlueBits = 0;
			pfd.cBlueShift = 0;
			pfd.cAlphaBits = 0;
			pfd.cAlphaShift = 0;
			pfd.cAccumBits = 0;
			pfd.cAccumRedBits = 0;
			pfd.cAccumGreenBits = 0;
			pfd.cAccumBlueBits = 0;
			pfd.cAccumAlphaBits = 0;
			pfd.cDepthBits = 16;
			pfd.cStencilBits = 0;
			pfd.cAuxBuffers = 0;
			pfd.iLayerType = (byte)Gdi.PFD_MAIN_PLANE;
			pfd.bReserved = 0;
			pfd.dwLayerMask = 0;
			pfd.dwVisibleMask = 0;
			pfd.dwDamageMask = 0;

			// Attempt to get a device contexts
			hDC = User.GetDC(renderPanel.Handle);
			if(hDC == IntPtr.Zero) {
				MessageBox.Show("Can't Create A GL Device Context.", "ERROR", 
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// Find pixelformat
			int pixelFormat = Gdi.ChoosePixelFormat(hDC, ref pfd);
			if(pixelFormat == 0) {
				deleteRenderContext();
				MessageBox.Show("Can't Find A Suitable PixelFormat.", "ERROR", 
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// Set the pixelformat
			if(!Gdi.SetPixelFormat(hDC, pixelFormat, ref pfd)) {
				deleteRenderContext();
				MessageBox.Show("Can't Set The PixelFormat.", "ERROR", 
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// Create rendering context
			hRC = Wgl.wglCreateContext(hDC);
			if(hRC == IntPtr.Zero) {
				deleteRenderContext();
				MessageBox.Show("Can't Create A GL Rendering Context.", "ERROR", 
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			// Make the device and rendering context active
			if(!Wgl.wglMakeCurrent(hDC, hRC)) {
				deleteRenderContext();
				MessageBox.Show("Can't Activate The GL Rendering Context.", "ERROR", 
					MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}
			
			return true;
		}

		#endregion

		#region Delete Rendering Context

		private void deleteRenderContext()
		{
			// Rendering context
			if(hRC != IntPtr.Zero) {

				// Release device and rendering context
				if(!Wgl.wglMakeCurrent(IntPtr.Zero, IntPtr.Zero)) {
					MessageBox.Show("Release Of DC And RC Failed.", "SHUTDOWN ERROR", 
						MessageBoxButtons.OK, MessageBoxIcon.Error);
				}

				// Delete rendering context
				if(!Wgl.wglDeleteContext(hRC)) {
					MessageBox.Show("Release Rendering Context Failed.", "SHUTDOWN ERROR",
						MessageBoxButtons.OK, MessageBoxIcon.Error);
				}

				// Set rendering context to Null
				hRC = IntPtr.Zero;
			}

			// Device context
			if(hDC != IntPtr.Zero) {

				// Do we have a window handle
				if(this.Handle != IntPtr.Zero) {

					// Release the device context
					if(!User.ReleaseDC(this.Handle, hDC)) {
						MessageBox.Show("Release Device Context Failed.", "SHUTDOWN ERROR",
							MessageBoxButtons.OK, MessageBoxIcon.Error);
					}
				}				

				// Set device context to Null
				hDC = IntPtr.Zero;
			}
		}

		#endregion

		#region Initialize OpenGL

		private bool initOpenGL() {
			Gl.glShadeModel(Gl.GL_SMOOTH);	// Enable Smooth Shading
			
			Gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);	// Clear Color
			Gl.glClearDepth(1);				// Depth Buffer Setup
			
			Gl.glEnable(Gl.GL_DEPTH_TEST);	// Enables Depth Testing
			Gl.glDepthFunc(Gl.GL_LEQUAL);	// The Type Of Depth Testing To Do
			
			Gl.glHint(Gl.GL_PERSPECTIVE_CORRECTION_HINT, Gl.GL_NICEST);	// Really Nice Perspective Calculations

			Gl.glEnable(Gl.GL_LIGHTING);
			Gl.glEnable(Gl.GL_LIGHT0);

			Gl.glColor3d(1.0, 1.0, 1.0);

			//Gl.glCullFace(Gl.GL_FRONT_AND_BACK);

			return true;
		}

		#endregion

		#region Set View Port

		private void setViewPort(int x, int y, int width, int height) {
			// Prevent divide by zero
			if (height == 0) {
				height = 1;
			}

			Gl.glViewport(x, y, width, height);
			Gl.glMatrixMode(Gl.GL_PROJECTION);
			Gl.glLoadIdentity();
			Glu.gluPerspective(45, width/(double)height, 0.1, 1000);
			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glLoadIdentity();
		}

		#endregion

		#region Render Frame

		private bool renderFrame() {
			Gl.glClear(Gl.GL_COLOR_BUFFER_BIT | Gl.GL_DEPTH_BUFFER_BIT);
			Gl.glLoadIdentity();
			Gl.glColor3d(1.0, 1.0, 1.0);

			float[] lightPos = new float[3];
			lightPos[0] = 0.0F; lightPos[1] = 1.0F; lightPos[2] = 2.0F;
			Gl.glLightfv(Gl.GL_LIGHT0, Gl.GL_POSITION, lightPos);

			camera.lookThrough();

			if(doRenderSkin) {
				renderSkin();
			}
			if(doRenderWireframe) {
				renderWireframe();
			}
			if(doRenderSkeleton) {
				renderSkeleton();
			}

			return true;
		}

		#endregion

		#region Render Methods

		private void renderSkin() {
			Mesh mesh = animator.getMesh();

			Gl.glEnableClientState(Gl.GL_VERTEX_ARRAY);
			Gl.glVertexPointer(3, Gl.GL_DOUBLE, 0, mesh.getVertexArray());
			Gl.glEnableClientState(Gl.GL_NORMAL_ARRAY);
			Gl.glNormalPointer(Gl.GL_DOUBLE, 0, mesh.getNormalArray());

			Gl.glCullFace(Gl.GL_BACK);
			Gl.glEnable(Gl.GL_CULL_FACE);

			Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_DIFFUSE, new float[] { 0.8f, 0.8f, 1.0f });
			Gl.glDrawArrays(Gl.GL_TRIANGLES, 0, mesh.getNumArrayElements());

			Gl.glDisableClientState(Gl.GL_VERTEX_ARRAY);
			Gl.glDisableClientState(Gl.GL_NORMAL_ARRAY);
		}

		private void renderWireframe() {
			Mesh mesh = animator.getMesh();

			Gl.glEnableClientState(Gl.GL_VERTEX_ARRAY);
			Gl.glVertexPointer(3, Gl.GL_DOUBLE, 0, mesh.getVertexArray());
			Gl.glEnableClientState(Gl.GL_NORMAL_ARRAY);
			Gl.glNormalPointer(Gl.GL_DOUBLE, 0, mesh.getNormalArray());

			Gl.glDisable(Gl.GL_LIGHTING);
			
			int numElements = mesh.getNumArrayElements();
			for(int i=0; i<numElements; i+=3) {
				Gl.glDrawArrays(Gl.GL_LINE_LOOP, i, 3);
			}

			Gl.glEnable(Gl.GL_LIGHTING);

			Gl.glDisableClientState(Gl.GL_VERTEX_ARRAY);
			Gl.glDisableClientState(Gl.GL_NORMAL_ARRAY);
		}

		private void renderSkeleton() {
			Skeleton skeleton = animator.getMesh().getSkeleton();

			Gl.glDisable(Gl.GL_LIGHTING);

			uint numJoints =(uint)skeleton.getJoints().Count;
			for(uint i=0; i<numJoints; i++) {
				Gl.glPushMatrix();
				Matrix tranformation = skeleton.getJoint(i).getAbsoluteTransformation();
				Gl.glMultMatrixd(tranformation.getValues());
				
				// X Axis
				Gl.glBegin(Gl.GL_LINE_STRIP);
				Gl.glColor3d(1.0, 0.0, 0.0);
				Gl.glVertex3d(0.0, 0.0, 0.0);
				Gl.glVertex3d(0.5, 0.0, 0.0);
				Gl.glEnd();

				// Y Axis
				Gl.glBegin(Gl.GL_LINE_STRIP);
				Gl.glColor3d(0.0, 1.0, 0.0);
				Gl.glVertex3d(0.0, 0.0, 0.0);
				Gl.glVertex3d(0.0, 0.5, 0.0);
				Gl.glEnd();

				// Z Axis
				Gl.glBegin(Gl.GL_LINE_STRIP);
				Gl.glColor3d(0.0, 0.0, 1.0);
				Gl.glVertex3d(0.0, 0.0, 0.0);
				Gl.glVertex3d(0.0, 0.0, 0.5);
				Gl.glEnd();				
				
				Gl.glPopMatrix();
			}

			Gl.glEnable(Gl.GL_LIGHTING);
		}

		#endregion

		#region Event Handlers

		private void tickRefreshTimer(object sender, EventArgs e) {
			setViewPort(renderPanel.Left, renderPanel.Top, renderPanel.Width, renderPanel.Height);
			renderFrame();
			Gdi.SwapBuffers(hDC);
			
		}

		private void tickAnimationTimer(object sender, EventArgs e) {
			animator.updateAnimation();
		}

		private void valueChangedFPS(object sender, EventArgs e) {
			animationTimer.Interval = 1000/fpsTrackBar.Value;
		}

		private void hScrollBar_ValueChanged(object sender, EventArgs e) {
			camera.setYaw(hScrollBar.Value);
		}

		private void vScrollBar_ValueChanged(object sender, EventArgs e) {
			camera.setPitch(-vScrollBar.Value);
		}

		private void buttonClickPlayPause(object sender, EventArgs e) {
			if(animator.getRunning()) {
				animator.setRunning(false);
				buttonPlayPause.Text = "Play";
			}
			else {
				animator.setRunning(true);
				buttonPlayPause.Text = "Pause";
			}
		}

		private void buttonStop_Click(object sender, EventArgs e) {
			animator.setRunning(false);
			animator.getMesh().setBindPose();
			animator.resetAnimations();
			buttonPlayPause.Text = "Play";
		}


		private void checkBoxSkin_CheckedChanged(object sender, EventArgs e) {
			doRenderSkin = checkBoxSkin.Checked;
		}

		private void checkBoxWireframe_CheckedChanged(object sender, EventArgs e) {
			doRenderWireframe = checkBoxWireframe.Checked;
		}

		private void checkBoxSkeleton_CheckedChanged(object sender, EventArgs e) {
			doRenderSkeleton = checkBoxSkeleton.Checked;
		}
			
		private void checkBoxWave_CheckedChanged(object sender, EventArgs e) {
			if(checkBoxWave.Checked) {
				animator.addDominantAnimation((Animation)animations[1]);
			}
			else {
				animator.removeDominantAnimation((Animation)animations[1]);
			}
		}

		private void comboBoxMeshes_SelectedIndexChanged(object sender, EventArgs e) {
			Mesh currentMesh = ((MeshData)meshes[comboBoxMeshes.SelectedIndex]).createInstance();
			currentMesh.transformVertices();
			animator.setMesh(currentMesh);
		}

		#endregion				
			
	}
}