using System;
using System.Collections;
using System.Text;
using Util;

namespace SkeletalAnimation {

	struct MergeAnimation {
		public Animation animation;
		public double weight;
	}
	
	/**
	 * This class is responsible for animating the mesh. It only supports
	 * one mesh for now.
	 */
	public class Animator {

		#region Member Variables

		private bool running;
		
		private Mesh mesh;

		private ArrayList mergeAnimations;	// Animations that blend together
		private ArrayList dominantAnimations;	// Animations that override merged animations

		#endregion

		#region Getters And Setters

		public void setMesh(Mesh m) { mesh = m; }
		public Mesh getMesh() { return mesh; }

		public void setRunning(bool run) { running = run; }
		public bool getRunning() { return running; }

		#endregion

		#region Constructors

		/**
		 * Default constructor
		 */
		public Animator() {
			running = false;
			mergeAnimations = new ArrayList();
			dominantAnimations = new ArrayList();
		}

		#endregion

		#region Methods

		public void addMergeAnimation(Animation animation, double weight) {
			MergeAnimation mergeAnim = new MergeAnimation();
			mergeAnim.animation = animation;
			mergeAnim.weight = weight;
			
			mergeAnimations.Add(animation);
		}

		public void addDominantAnimation(Animation animation) {
			dominantAnimations.Add(animation);
		}

		public void removeDominantAnimation(Animation animation) {
			dominantAnimations.Remove(animation);
		}

		/** 
		 * Reset all the current animations ie set them to frame 1
		 */
		public void resetAnimations() {

			

			// Loop through all the merged animations
			for(int animationIndex=0; animationIndex<mergeAnimations.Count; animationIndex++) {
				Animation animation = (Animation)mergeAnimations[animationIndex];
				animation.setCurrentFrame(1);
			}

			// Loop through all the dominant animations
			for(int animationIndex=0; animationIndex<dominantAnimations.Count; animationIndex++) {
				Animation animation = (Animation)mergeAnimations[animationIndex];
				animation.setCurrentFrame(1);
			}
		}

		public void updateAnimation() {
			if(!running) {
				return;
			}

			// Loop through all the merged animations
			for(int animationIndex=0; animationIndex<mergeAnimations.Count; animationIndex++) {
				Animation animation = (Animation)mergeAnimations[animationIndex];
				
				// If we've reached the end of the animation loop
				if(animation.getCurrentFrame() == animation.getNumFrames()-1) {
					animation.setCurrentFrame(1);
				}
				else {
					animation.setCurrentFrame(animation.getCurrentFrame()+1);
				}

				int numTracks = animation.getNumTracks();
				for(int i=0; i<numTracks; i++) {
					AnimationTrack track = animation.getTrack(i);

					// If the current frame is a keyframe 
					if(animation.isKeyFrame(i, animation.getCurrentFrame())) {
						KeyFrame keyFrame = animation.getKeyFrame(i, animation.getCurrentFrame());
						mesh.getSkeleton().getJoint(track.jointId).setRelativeRotation(keyFrame.rotation);
					}

					// If we need to interpolate
					else {
						KeyFrame nextKey = animation.getNextKeyFrame(i, animation.getCurrentFrame());
						KeyFrame prevKey = animation.getPrevKeyFrame(i, animation.getCurrentFrame());

						double keySpan;
						if(nextKey.frame < prevKey.frame) {
							keySpan = animation.getNumFrames() - prevKey.frame + nextKey.frame;
						}
						else {
							keySpan = nextKey.frame - prevKey.frame;
						}

						double keyOff;
						if(animation.getCurrentFrame() < prevKey.frame) {
							keyOff = animation.getNumFrames() - prevKey.frame + animation.getCurrentFrame();
						}
						else {
							keyOff = animation.getCurrentFrame() - prevKey.frame;
						}

						Quaternion rotation = Quaternion.slerpNoInvert(prevKey.rotation, nextKey.rotation, keyOff/keySpan);
						mesh.getSkeleton().getJoint(track.jointId).setRelativeRotation(rotation);
					}
				}
			}

			// Loop through all the dominant animations
			for(int animationIndex=0; animationIndex<dominantAnimations.Count; animationIndex++) {
				Animation animation = (Animation)dominantAnimations[animationIndex];

				// If we've reached the end of the animation loop
				if(animation.getCurrentFrame() == animation.getNumFrames()-1) {
					animation.setCurrentFrame(1);
				}
				else {
					animation.setCurrentFrame(animation.getCurrentFrame()+1);
				}

				int numTracks = animation.getNumTracks();
				for(int i=0; i<numTracks; i++) {
					AnimationTrack track = animation.getTrack(i);

					// If the current frame is a key frame
					if(animation.isKeyFrame(i, animation.getCurrentFrame())) {
						KeyFrame keyFrame = animation.getKeyFrame(i, animation.getCurrentFrame());
						mesh.getSkeleton().getJoint(track.jointId).setRelativeRotation(keyFrame.rotation);
					}

					// If we need to interpolate
					else {
						KeyFrame nextKey = animation.getNextKeyFrame(i, animation.getCurrentFrame());
						KeyFrame prevKey = animation.getPrevKeyFrame(i, animation.getCurrentFrame());

						double keySpan;
						if(nextKey.frame < prevKey.frame) {
							keySpan = animation.getNumFrames() - prevKey.frame + nextKey.frame;
						}
						else {
							keySpan = nextKey.frame - prevKey.frame;
						}

						double keyOff;
						if(animation.getCurrentFrame() < prevKey.frame) {
							keyOff = animation.getNumFrames() - prevKey.frame + animation.getCurrentFrame();
						}
						else {
							keyOff = animation.getCurrentFrame() - prevKey.frame;
						}

						Quaternion rotation = Quaternion.slerpNoInvert(prevKey.rotation, nextKey.rotation, keyOff/keySpan);
						mesh.getSkeleton().getJoint(track.jointId).setRelativeRotation(rotation);
					}
				}
			}

			// Update the absolutes
			uint numJoints = (uint)mesh.getSkeleton().getJoints().Count;
			for(uint i=0; i<numJoints; i++) {
				mesh.getSkeleton().getJoint(i).updateAbsolutes();
			}

			mesh.transformVertices();
		}

		#endregion
	}
}
