using System;
using System.Collections;
using System.Text;
using Util;

namespace SkeletalAnimation {
	
	public class Mesh {

		#region Member Variables

		private Skeleton skeleton;
		private MeshData meshData;

		private double[] vertices;
		private double[] normals;
		
		private int numArrayElements;
		private double[] vertexArray;
		private double[] normalArray;
		private double[] uvArray;

		#endregion

		#region Constructors

		/**
		 * Default Constructor
		 */
		public Mesh() {
			
		}

		#endregion

		#region Transform Vertices

		/**
		 * Transform the vertices
		 */
		public void transformVertices() {
						
			// Find the joint transformations and rotations
			ArrayList joints = skeleton.getJoints();
			Matrix[] transformations = new Matrix[joints.Count];
			Matrix[] rotations = new Matrix[joints.Count];
			for(int i=0; i<joints.Count; i++) {
				Joint joint = (Joint)joints[i];
				
				// Transformed * Original.. This seems backwards to me....
				transformations[i] = joint.getAbsoluteTransformation() * 
					joint.getJointData().getAbsoluteTransformationInverse();

				rotations[i] = joint.getAbsoluteRotation().toMatrix() *
					joint.getJointData().getAbsoluteRotation().toMatrix().getTransposed();
			}

			// Transform the vertices
			double[] orgVertices = meshData.getVertices();
			for(uint vertexIndex=0; vertexIndex<meshData.getInfluences().Length; vertexIndex++) {

				ArrayList vertexInfluences = meshData.getInfluences(vertexIndex);
				int influenceIndex = 0;
				BoneInfluence influence = (BoneInfluence)vertexInfluences[influenceIndex];
				Matrix jointTrans = transformations[influence.boneId] * influence.weight;
				Matrix totalTrans = jointTrans;
				for(influenceIndex=1; influenceIndex<vertexInfluences.Count; influenceIndex++) {
					influence = (BoneInfluence)vertexInfluences[influenceIndex];
					jointTrans = transformations[influence.boneId] * influence.weight;
					totalTrans = totalTrans + jointTrans;
				}
				Vector orgVertex = new Vector(orgVertices[vertexIndex*3], orgVertices[vertexIndex*3+1], orgVertices[vertexIndex*3+2]); 
				Vector vertex = totalTrans * orgVertex;
				vertices[vertexIndex*3] = vertex[0]; 
				vertices[vertexIndex*3+1] = vertex[1]; 
				vertices[vertexIndex*3+2] = vertex[2];
			}

			// Transform the normals 
			double[] orgNormals = meshData.getNormals();
			uint[] normalIndices = meshData.getNormalIndices();
			uint[] normalToVertex = meshData.getNormalToVertex();
			uint[] vertexIndices = meshData.getVertexIndices();
			int numNormals = normals.Length / 3;

			for(int normalIndex=0; normalIndex<numNormals; normalIndex++) {
				
				// Find the rotation of the corresponding vertex
				uint vertexIndex = normalToVertex[normalIndex];
				ArrayList vertexInfluences = meshData.getInfluences(vertexIndex);
				int influenceIndex = 0;
				BoneInfluence influence = (BoneInfluence)vertexInfluences[influenceIndex];
				Matrix jointRot = rotations[influence.boneId] * influence.weight;
				Matrix totalRot = jointRot;
				for(influenceIndex=1; influenceIndex<vertexInfluences.Count; influenceIndex++) {
					influence = (BoneInfluence)vertexInfluences[influenceIndex];
					jointRot = rotations[influence.boneId] * influence.weight;
					totalRot = totalRot + jointRot;
				}

				Vector orgNormal = new Vector(orgNormals[normalIndex*3], orgNormals[normalIndex*3+1], orgNormals[normalIndex*3+2]);
				Vector normal = totalRot * orgNormal;
				normals[normalIndex*3] = normal[0];
				normals[normalIndex*3+1] = normal[1];
				normals[normalIndex*3+2] = normal[2];
			}
			
			// Fill in vertex arrays
			for(int index=0; index<numArrayElements; index++) {
				vertexArray[index*3] = vertices[vertexIndices[index]*3];
				vertexArray[index*3+1] = vertices[vertexIndices[index]*3+1];
				vertexArray[index*3+2] = vertices[vertexIndices[index]*3+2];
				
				normalArray[index*3] = normals[normalIndices[index]*3];
				normalArray[index*3+1] = normals[normalIndices[index]*3+1];
				normalArray[index*3+2] = normals[normalIndices[index]*3+2];
			}			
		}

		#endregion

		#region Methods

		/**
		 * Create Vertex Arrays. This method creates the arrays for vertex array rendering
		 */
		public void createVertexArrays(int numElements) {
			vertexArray = new double[numElements * 3];
			normalArray = new double[numElements * 3];
			uvArray = new double[numElements * 2];
			numArrayElements = numElements;
		}

		/**
		 * Fill in UV array. This can be done once since UV's aren't transformed by the skeleton
		 */
		public void fillInUVArray() {
			uint[] uvIndices = meshData.getUVIndices();
			double[] uvs = meshData.getUVs();
			for(int index=0; index<numArrayElements; index++) {
				uvArray[index*2] = uvs[uvIndices[index]*2];
				uvArray[index*2+1] = uvs[uvIndices[index]*2+1];
			}
		}

		/**
		 * Set the mesh to the bind pose
		 */
		public void setBindPose() {
			skeleton.setBindPose();
			
			// Transform the vertices
			transformVertices();
		}

		#endregion

		#region Getters & Setters

		public void setMeshData(MeshData md) { meshData = md; }
		public MeshData getMeshData() { return meshData; }

		public void setSkeleton(Skeleton skel) { skeleton = skel; }
		public Skeleton getSkeleton() { return skeleton; }

		public void setVertices(double[] verts) { vertices = verts; }
		public double[] getVertices() { return vertices; }

		public void setNormals(double[] norms) { normals = norms; }
		public double[] getNormals() { return normals; }

		//public void setUVs(double[] uv) { uvs = uv; }
		//public double[] getUVs() { return uvs; }

		public int getNumArrayElements() { return numArrayElements; }

		public double[] getVertexArray() { return vertexArray; }
		public double[] getNormalArray() { return normalArray; }
		public double[] getUVArray() { return uvArray; }

		#endregion
	}
}
