using System;
using System.Collections;
using System.Text;
using System.IO;
using System.Globalization;
using Util;

namespace SkeletalAnimation {
	/**
	 * Rite Ascii Model Loader
	 */
	public class RAMLoader {

        
		#region Read RAM file

		public MeshData readRAM(string filename) {
			StreamReader fileStream = new StreamReader(filename);
			String fileData = fileStream.ReadToEnd();
			fileStream.Close();

			char[] delimiters = new char[4];
			delimiters[0] = ' ';
			delimiters[1] = '\t';
			delimiters[2] = '\r';
			delimiters[3] = '\n';
						
			String[] substrings = fileData.Split(delimiters, StringSplitOptions.RemoveEmptyEntries);
			
			// Number format
			NumberFormatInfo numberFormat = new NumberFormatInfo();
			numberFormat.NumberDecimalSeparator = ".";

			// The new mesh
			MeshData newMesh = new MeshData();
			
			uint index = 0;

			// Extract joints
			if(substrings[index++] == "joints") {
				uint numJoints = UInt32.Parse(substrings[index++]);
				
				for(uint i=0; i<numJoints; i++) {
					index++; // Skip the index

					JointData newJoint = new JointData();

					delimiters = new char[1];
					delimiters[0] = '"';
					String[] nameArray = substrings[index++].Split(delimiters, StringSplitOptions.RemoveEmptyEntries);
					newJoint.setName(nameArray[0]);
					
					//	Read parent joint 
					index++;	// Skip parent keyword
					newJoint.setParentJoint(Int32.Parse(substrings[index++]));

					// Read orientation
					index++;	// Skip orientation keyword
					Quaternion orientation = new Quaternion(Double.Parse(substrings[index++], numberFormat),
						Double.Parse(substrings[index++], numberFormat),
						Double.Parse(substrings[index++], numberFormat),
						Double.Parse(substrings[index++], numberFormat));
					newJoint.setOrientation(orientation);

					// Read translation
					index++;	// Skip translation keyword
					Vector relativeTranslation = new Vector(Double.Parse(substrings[index++], numberFormat), 
						Double.Parse(substrings[index++], numberFormat), 
						Double.Parse(substrings[index++], numberFormat));
					newJoint.setRelativeTranslation(relativeTranslation);

					// Read rotation
					index++;	// Skip rotation keyword
					Quaternion relativeRotation = new Quaternion(Double.Parse(substrings[index++], numberFormat),
						Double.Parse(substrings[index++], numberFormat),
						Double.Parse(substrings[index++], numberFormat),
						Double.Parse(substrings[index++], numberFormat));
					newJoint.setRelativeRotation(relativeRotation);

					// Add the joint to the new mesh skeleton
					newJoint.setSkeletonData(newMesh.getSkeletonData());
					newJoint.updateAbsolutes();
					newMesh.getSkeletonData().getJointData().Add(newJoint);
				}
			}
			
			// Extract mesh name
			if(substrings[index++] == "mesh") {
				delimiters = new char[1];
				delimiters[0] = '"';
				
				String[] nameArray = substrings[index++].Split(delimiters, StringSplitOptions.RemoveEmptyEntries);
				newMesh.setName(nameArray[0]);
			}
			
			// Extract vertices
			uint numVertices = 0;
			if(substrings[index++] == "vertices") {
				numVertices = UInt32.Parse(substrings[index++]);
				double[] vertices = new double[numVertices*3];
				
				for(uint i=0; i<numVertices; i++) {
					index++;    // Skip the index
					
					vertices[i*3] = Double.Parse(substrings[index++], numberFormat);
					vertices[i*3+1] = Double.Parse(substrings[index++], numberFormat);
					vertices[i*3+2] = Double.Parse(substrings[index++], numberFormat);
				}
				// Insert into new mesh
				newMesh.setVertices(vertices);
			}
			
			// Extract normals
			uint numNormals = 0;
			if(substrings[index++] == "normals") {
				numNormals = UInt32.Parse(substrings[index++]);
				double[] normals = new double[numNormals*3];
				
				for(uint i=0; i<numNormals; i++) {
					
					index++;    // Skip the index
					
					normals[i*3] = Double.Parse(substrings[index++], numberFormat);
					normals[i*3+1] = Double.Parse(substrings[index++], numberFormat);
					normals[i*3+2] = Double.Parse(substrings[index++], numberFormat);
				}
				// Insert into new mesh
				newMesh.setNormals(normals);
			}
			
			// Extract uvs
			if(substrings[index++] == "uvs") {
				uint numUvs = UInt32.Parse(substrings[index++]);
				double[] uvs = new double[numUvs*2];
				
				for(uint i=0; i<numUvs; i++) {
					index++;    // Skip the index
					
					uvs[i*2] = Double.Parse(substrings[index++], numberFormat);
					uvs[i*2+1] = Double.Parse(substrings[index++], numberFormat);
				}
				
				// Insert into new mesh
				newMesh.setUVs(uvs);
			}
			
			// Extract triangles and tie together normals and vertices
			if(substrings[index++] == "triangles") {
				
				uint numTriangles = UInt32.Parse(substrings[index++]);
				uint[] vertexIndices = new uint[numTriangles*3];
				uint[] normalIndices = new uint[numTriangles*3];
				uint[] normalToVertex = new uint[numNormals];
				uint[] uvIndices = new uint[numTriangles*3];
				
				for(uint i=0; i<numTriangles; i++) {
					index++;    // Skip the index
					index++;	// Skip the shader

					vertexIndices[i*3] = UInt32.Parse(substrings[index++]);
					vertexIndices[i*3+1] = UInt32.Parse(substrings[index++]);
					vertexIndices[i*3+2] = UInt32.Parse(substrings[index++]);

					normalIndices[i*3] = UInt32.Parse(substrings[index++]);
					normalIndices[i*3+1] = UInt32.Parse(substrings[index++]);
					normalIndices[i*3+2] = UInt32.Parse(substrings[index++]);

					// Relate normal to vertex
					normalToVertex[normalIndices[i*3]] = vertexIndices[i*3];
					normalToVertex[normalIndices[i*3+1]] = vertexIndices[i*3+1];
					normalToVertex[normalIndices[i*3+2]] = vertexIndices[i*3+2];

					uvIndices[i*3] = UInt32.Parse(substrings[index++]);
					uvIndices[i*3+1] = UInt32.Parse(substrings[index++]);
					uvIndices[i*3+2] = UInt32.Parse(substrings[index++]);
				}
				newMesh.setVertexIndices(vertexIndices);
				newMesh.setNormalIndices(normalIndices);
				newMesh.setNormalToVertex(normalToVertex);
				newMesh.setUVIndices(uvIndices);
			}

			// Extract weights
			if(substrings[index++] == "weights") {
				uint numWeights = UInt32.Parse(substrings[index++]);
				
				newMesh.setInfluences(new ArrayList[numVertices]);
				for(int i=0; i<numVertices; i++) {
					newMesh.getInfluences()[i] = new ArrayList();
				}

				for(int i=0; i<numWeights; i++) {
					index++;	// Skip the index

					uint vertexIndex = UInt32.Parse(substrings[index++]);
					uint boneIndex = UInt32.Parse(substrings[index++]);
					index++;	// Skip the joint name
					double weight = Double.Parse(substrings[index++], numberFormat);

					// Insert the new joint influence into the new mesh
					BoneInfluence newInfluence = new BoneInfluence();
					newInfluence.boneId = boneIndex;
					newInfluence.weight = weight;
					newMesh.getInfluences()[vertexIndex].Add(newInfluence);
				}								
			}
			return newMesh;
		}
		
		#endregion
	}
}
